--Json = require("Json")
--io = require("io")

local function sliceImage(data, x, y, w, h)
    local slice = love.image.newImageData(w, h)
    slice:paste(data, 0, 0, x, y, w, h)
    return love.graphics.newImage(slice)
end

local function resetAnimation()
    Player.stages = 0

    Player.left_leg.x = 0
    Player.left_leg.y = 22

    Player.right_leg.x = 16
    Player.right_leg.y = 22

    Player.head.x = 2
    Player.head.y = -16

    Player.body.x = 0
    Player.body.y = 0

    Player.left_arm.x = -10
    Player.left_arm.y = 10

    Player.right_arm.x = 30
    Player.right_arm.y = 10
end

local function checkCollision(x1, y1, w1, h1, x2, y2, w2, h2)
    return x1 < x2+w2 and
           x2 < x1+w1 and
           y1 < y2+h2 and
           y2 < y1+h1
end

local function drawPlayer(x, y)
    Objects.new("p_right_leg", Player.right_leg.sprite, x+Player.right_leg.x, y+Player.right_leg.y, Player.right_leg.width, Player.right_leg.height, false, 1, 0, 1, 1)
    Objects.new("p_left_leg", Player.left_leg.sprite, x+Player.left_leg.x, y+Player.left_leg.y, Player.left_leg.width, Player.left_leg.height, false, 1, 0, 1, 1)

    Objects.new("p_body", Player.body.sprite, x+Player.body.x, y+Player.body.y, Player.body.width, Player.body.height, true, 2, 0, 1, 1)
    Objects.new("p_head", Player.head.sprite, x+Player.head.x, y+Player.head.y, Player.head.width, Player.head.height, true, 3, 0, 1, 1)

    Objects.new("p_right_arm", Player.right_arm.sprite, x+Player.right_arm.x, y+Player.right_arm.y, Player.right_arm.width, Player.right_arm.height, false, 1, 0, 1, 1)
    Objects.new("p_left_arm", Player.left_arm.sprite, x+Player.left_arm.x, y+Player.left_arm.y, Player.left_arm.width, Player.left_arm.height, false, 1, 0, 1, 1)
end

local function updatePlayer(x, y, s, st)
    local waypoint = 9
    local balance = 3

    local increasement = 1

    if st=="idle" then
        if Player.stages <= waypoint then
            if Player.stages % balance == 0 then
                Player.head.y = Player.head.y + increasement
                Player.body.y = Player.body.y + increasement

                Player.right_arm.y = Player.right_arm.y + increasement
                Player.left_arm.y = Player.left_arm.y + increasement
            end
            Player.stages = Player.stages + 1
        elseif Player.stages <= waypoint * 2 then
            if Player.stages % balance == 0 then
                Player.head.y = Player.head.y - increasement
                Player.body.y = Player.body.y - increasement

                Player.right_arm.y = Player.right_arm.y - increasement
                Player.left_arm.y = Player.left_arm.y - increasement
            end
            Player.stages = Player.stages + 1
        else
            Player.head.y = Player.head.y - increasement
            Player.body.y = Player.body.y - increasement

            Player.right_arm.y = Player.right_arm.y - increasement
            Player.left_arm.y = Player.left_arm.y - increasement

            Player.stages = 0
        end
    elseif st=="move" then
        if Player.stages <= waypoint then
            if Player.stages == 0 then
                Player.head.y = -13

                Player.right_leg.y = 22
                Player.left_leg.y = 19
            elseif Player.stages % balance == 0 then
                Player.head.y = Player.head.y - increasement

                Player.right_leg.y = Player.right_leg.y - increasement
                Player.left_leg.y = Player.left_leg.y + increasement
            end
            Player.stages = Player.stages + 1
        elseif Player.stages <= waypoint * 2 then
            if Player.stages % balance == 0 then
                Player.head.y = Player.head.y + increasement

                Player.right_leg.y = Player.right_leg.y + increasement
                Player.left_leg.y = Player.left_leg.y - increasement
            end
            Player.stages = Player.stages + 1
        else
            Player.head.y = Player.head.y + increasement

            Player.right_leg.y = Player.right_leg.y + increasement
            Player.left_leg.y = Player.left_leg.y - increasement

            Player.stages = 0
        end
    end

    local rotatePlus = 30
    if s == 0 then
        Objects.set(Objects.find("p_right_leg"), Player.right_leg.sprite, x+Player.right_leg.x, y+Player.right_leg.y, false, 1, 0, 1, 1)
        Objects.set(Objects.find("p_left_leg"), Player.left_leg.sprite, x+Player.left_leg.x, y+Player.left_leg.y, false, 1, 0, 1, 1)

        Objects.set(Objects.find("p_body"), Player.body.sprite, x+Player.body.x, y+Player.body.y, true, 2, 0, 1, 1)
        Objects.set(Objects.find("p_head"), Player.head.sprite, x+Player.head.x, y+Player.head.y, true, 3, 0, 1, 1)

        Objects.set(Objects.find("p_right_arm"), Player.right_arm.sprite, x+Player.right_arm.x, y+Player.right_arm.y, false, 1, 0, 1, 1)
        Objects.set(Objects.find("p_left_arm"), Player.left_arm.sprite, x+Player.left_arm.x, y+Player.left_arm.y, false, 1, 0, 1, 1)
    else
        Objects.set(Objects.find("p_right_leg"), Player.right_leg.sprite, x-Player.right_leg.x+rotatePlus, y+Player.right_leg.y, false, 1, 0, -1, 1)
        Objects.set(Objects.find("p_left_leg"), Player.left_leg.sprite, x-Player.left_leg.x+rotatePlus, y+Player.left_leg.y, false, 1, 0, -1, 1)

        Objects.set(Objects.find("p_body"), Player.body.sprite, x-Player.body.x+rotatePlus, y+Player.body.y, true, 2, 0, -1, 1)
        Objects.set(Objects.find("p_head"), Player.head.sprite, x-Player.head.x+rotatePlus, y+Player.head.y, true, 3, 0, -1, 1)

        Objects.set(Objects.find("p_right_arm"), Player.right_arm.sprite, x-Player.right_arm.x+rotatePlus, y+Player.right_arm.y, false, 1, 0, -1, 1)
        Objects.set(Objects.find("p_left_arm"), Player.left_arm.sprite, x-Player.left_arm.x+rotatePlus, y+Player.left_arm.y, false, 1, 0, -1, 1)
    end
end

local function drawRoom(name, tileset, x, y, w, h)
    local width = w * 64
    local height = h * 64

    Objects.new(name .. "_left_up", tileset.left_up, x, y, tileset.left_up:getWidth(), tileset.left_up:getHeight(), true, 4, 0, 1, 1)
    Objects.new(name .. "_right_up", tileset.right_up, x+width, y, tileset.right_up:getWidth(), tileset.right_up:getHeight(), true, 4, 0, 1, 1)

    for i = 1, math.abs(w-1) do
        Objects.new(name .. "_middle_up_" .. i, tileset.middle_up, i*64+x, y, tileset.middle_up:getWidth(), tileset.middle_up:getHeight(), true, 4, 0, 1, 1)
    end

    Objects.new(name .. "_left_down", tileset.left_down, x, y+height, tileset.left_down:getWidth(), tileset.left_down:getHeight(), true, 4, 0, 1, 1)
    Objects.new(name .. "_right_down", tileset.right_down, x+width, y+height, tileset.right_down:getWidth(), tileset.right_down:getHeight(), true, 4, 0, 1, 1)

    for i = 1, math.abs(h-1) do
        Objects.new(name .. "_middle_left_" .. i, tileset.middle_left, x, i*64+y, tileset.middle_left:getWidth(), tileset.middle_left:getHeight(), true, 4, 0, 1, 1)
    end
    for i = 1, math.abs(h-1) do
        Objects.new(name .. "_middle_right_" .. i, tileset.middle_right, x+width, i*64+y, tileset.middle_right:getWidth(), tileset.middle_right:getHeight(), true, 4, 0, 1, 1)
    end
    for i = 1, math.abs(w-1) do
        Objects.new(name .. "_middle_down_" .. i, tileset.middle_down, i*64+x, height + y, tileset.middle_down:getWidth(), tileset.middle_down:getHeight(), true, 4, 0, 1, 1)
    end
    for i = 1, math.abs(w-1) do
        for j = 1, math.abs(h-1) do
            Objects.new(name .. "_floor_" .. i .. j, tileset.floor, x+i*64, y+j*64, tileset.floor:getWidth(), tileset.floor:getHeight(), false, 0, 0, 1, 1)
        end
    end
end

local function loadPlayers()
    local players = {}
    local files = love.filesystem.getDirectoryItems("content/players/")

    local id = 0

    for key, file in pairs(files) do
        io.input("content/players/"..file)
        local text = io.read("*all")
        local json = Json.decode(text)
        local data = love.data.decode("data", "base64", json.sprite)
        local image = love.image.newImageData(data)

        if json.speed + json.melee + json.ranged == 7 then
            local player = {}
            player.name = json.name

            player.speed = json.speed
            player.melee = json.melee
            player.ranged = json.ranged

            local head = sliceImage(image, 26, 6, 22, 18)
            local body = sliceImage(image, 0, 0, 26, 24)

            local left_arm = sliceImage(image, 58, 16, 8, 8)
            local right_arm = sliceImage(image, 58, 16, 8, 8)

            local left_leg = sliceImage(image, 48, 10, 10, 14)
            local right_leg = sliceImage(image, 48, 10, 10, 14)

            player.head = {sprite=head, x=2, y=-16, width=head:getWidth(), height=head:getHeight()}
            player.body = {sprite=body, x=0, y=0, width=body:getWidth(), height=body:getHeight()}
            player.left_arm = {sprite=left_arm, x=-10, y=10, width=left_arm:getWidth(), height=left_arm:getHeight()}
            player.right_arm = {sprite=right_arm, x=30, y=10, width=right_arm:getWidth(), height=right_arm:getHeight()}
            player.left_leg = {sprite=left_leg, x=0, y=22, width=left_leg:getWidth(), height=left_leg:getHeight()}
            player.right_leg = {sprite=right_leg, x=16, y=22, width=right_leg:getWidth(), height=right_leg:getHeight()}

            player.stages = 0

            players[id] = player
            id = id + 1
        else
            love.window.showMessageBox("Unbalanced character!", "Character " .. json.name .. " is unbalanced, so you can't use him", "warning")
        end
    end

    return players
end

local function loadTilesets()
    local tilesets = {}
    local files = love.filesystem.getDirectoryItems("Content/Tilesets")

    local id = 0

    for key, file in ipairs(files) do
        io.input("Content/Tilesets/" .. file)
        local text = io.read("*all")
        local json = Json.decode(text)
        local data = love.data.decode("data", "base64", json.sprite)
        local image = love.image.newImageData(data)
        local tileset = {}

        tileset.left_up = sliceImage(image, 0, 0, 64, 64)
        tileset.middle_up = sliceImage(image, 64, 0, 64, 64)
        tileset.right_up = sliceImage(image, 128, 0, 64, 64)

        tileset.middle_left = sliceImage(image, 0, 64, 64, 64)
        tileset.floor = sliceImage(image, 64, 64, 64, 64)
        tileset.middle_right = sliceImage(image, 128, 64, 64, 64)

        tileset.left_down = sliceImage(image, 0, 128, 64, 64)
        tileset.middle_down = sliceImage(image, 64, 128, 64, 64)
        tileset.right_down = sliceImage(image, 128, 128, 64, 64)

        tilesets[id] = tileset
        id = id + 1
    end
    return tilesets
end

function love.load()
    love.window.setMode(0, 0, {msaa=16})
    love.window.setTitle("Froguelike")

    Side = 0
    State = "idle"

    Movement = {x=0, y=0}

    Rooms = {}
    Rooms.list = {}
    Rooms.new = function (tileset, x, y, w, h)
                            local name = "room_" .. #Rooms.list
                            drawRoom(name, tileset, x, y, w, h)
                            table.insert(Rooms.list, {name=name, x=x, y=y, width=w, height=h})
                         end
    Objects = {}
    Objects.list = {}
    Objects.new = function (name, sprite, x, y, width, height, solid, layer, rotation, scale_x, scale_y)
                     table.insert(Objects.list, {name=name, sprite=sprite, x=x, y=y, width=width, height=height,
                                                 solid=solid, layer=layer, rotation=rotation, scale_x=scale_x, scale_y=scale_y})
                     Objects.sort()
                  end
    Objects.sort = function ()
                      table.sort(Objects.list, function (a, b) return a.layer < b.layer end)
                   end
    Objects.set = function (id, sprite, x, y, solid, layer, rotation, scale_x, scale_y)
                     Objects.list[id].sprite = sprite
                     Objects.list[id].x = x
                     Objects.list[id].y = y
                     Objects.list[id].solid = solid
                     Objects.list[id].layer = layer
                     Objects.list[id].rotation = rotation
                     Objects.list[id].scale_x = scale_x
                     Objects.list[id].scale_y = scale_y
                  end
    Objects.find = function (name)
                      for k, object in ipairs(Objects.list) do
                         if object.name == name then
                            return k
                         end
                      end
                      return 0
                   end
    Players = loadPlayers()
    Tilesets = loadTilesets()

    Player = Players[0]
    Rooms.new(Tilesets[0], 0, 0, 20, 20)
    Rooms.new(Tilesets[0], 64*20 + 128, 0, 20, 20)
    drawPlayer(love.graphics.getWidth() / 2 + 58, love.graphics.getHeight() / 2 + 40)

    local head = Objects.list[Objects.find("p_head")]
    local body = Objects.list[Objects.find("p_body")]
    LeftCollider = {x=body.x - body.width/2, y=head.y, width=Player.speed*1.5, height=body.height + head.height}
    RightCollider = {x=body.x + body.width*1.5, y=head.y, width=Player.speed*1.5, height=body.height + head.height}
    UpCollider = {x=body.x, y=head.y - head.height/2, width=body.width, height=Player.speed*1.5}
    DownCollider = {x=body.x, y=body.y + body.height, width=body.width, height=Player.speed*1.5}

    LeftCollision = false
    RightCollision = false
    UpCollision = false
    DownCollision = false

    Debug = false
end

function love.update()
    updatePlayer(love.graphics.getWidth() / 2 + 58, love.graphics.getHeight() / 2 + 40, Side, State)

    local v = false
    local h = false
    if love.keyboard.isDown("a") and not LeftCollision then
        Movement.x = -Player.speed
        Side = 1
        h = true
    elseif love.keyboard.isDown("d") and not RightCollision then
        Movement.x = Player.speed
        Side = 0
        h = true
    else
        h = false
        Movement.x = 0
    end
    if love.keyboard.isDown("w") and not UpCollision then
        Movement.y = -Player.speed
        v = true
    elseif love.keyboard.isDown("s") and not DownCollision then
        LastMove = "down"
        Movement.y = Player.speed
        v = true
    else
        v = false
        Movement.y = 0
    end
    if not v and not h and State ~= "idle" then
        State = "idle"
        resetAnimation()
    elseif v and State ~= "move" or h and State ~= "move" then
        State = "move"
        resetAnimation()
    end

    for k, object in ipairs(Objects.list) do
        if object.name ~= "p_right_leg" and object.name ~= "p_left_leg" and object.name ~= "p_head" and object.name ~= "p_body"
        and object.name ~= "p_right_arm" and object.name ~= "p_left_arm" then
            if object.solid then
                --print(object.name .. ": " .. "x: " .. object.x .. ", y: " .. object.y .. "; width: " .. object.width .. ", height: " .. object.height .. ";")
                LeftCollision = checkCollision(LeftCollider.x, LeftCollider.y, LeftCollider.width, LeftCollider.height, object.x, object.y, object.width, object.height)
                RightCollision = checkCollision(RightCollider.x, RightCollider.y, RightCollider.width, RightCollider.height, object.x, object.y, object.width, object.height)
                UpCollision = checkCollision(UpCollider.x, UpCollider.y, UpCollider.width, UpCollider.height, object.x, object.y, object.width, object.height)
                DownCollision = checkCollision(DownCollider.x, DownCollider.y, DownCollider.width, DownCollider.height, object.x, object.y, object.width, object.height)
            end
            object.x = object.x - Movement.x
            object.y = object.y - Movement.y
        end
    end

    if love.keyboard.isDown("lctrl") and love.keyboard.isDown("f1") then
        Debug = not Debug
    end
end

function love.draw()
    love.graphics.setBackgroundColor(153/255, 50/255, 76/255)
    local scale = love.graphics.getWidth() / love.graphics.getHeight() / 2
    love.graphics.scale(scale, scale)

    for k, object in ipairs(Objects.list) do
        love.graphics.draw(object.sprite, object.x, object.y, object.rotation, object.scale_x, object.scale_y)
    end
    if Debug then
        love.graphics.rectangle("fill", LeftCollider.x, LeftCollider.y, LeftCollider.width, LeftCollider.height)
        love.graphics.rectangle("fill", RightCollider.x, RightCollider.y, RightCollider.width, RightCollider.height)

        love.graphics.rectangle("fill", UpCollider.x, UpCollider.y, UpCollider.width, UpCollider.height)
        love.graphics.rectangle("fill", DownCollider.x, DownCollider.y, DownCollider.width, DownCollider.height)
    end
end